import pytest
import xml.etree.ElementTree as ET


# Define a fixture to load the XML file
@pytest.fixture
def osmchange_xml_obj():
    with open("tests/data/test_osmchange.xml", "r") as fh:
        return ET.parse(fh)


# Path to the changeset XML file
@pytest.fixture
def osmchange_file_path():
    return "tests/data/test_osmchange.xml"


# Path to the augmented diff XML file
@pytest.fixture
def adiff_file_path():
    return "tests/data/test_adiff.xml"



from osmdiff import Node, AugmentedDiff, Relation, Way
from typing_extensions import assert_type


class TestAugmentedDiff:
    "tests for AugmentedDiff class"

    def test_init_augmenteddiff(self):
        "Test AugmentedDiff init"
        augmenteddiff = AugmentedDiff()
        assert_type(augmenteddiff, AugmentedDiff)
        assert_type(augmenteddiff.create, list)
        assert_type(augmenteddiff.modify, list)
        assert_type(augmenteddiff.delete, list)
        assert len(augmenteddiff.create) == 0
        assert len(augmenteddiff.modify) == 0
        assert len(augmenteddiff.delete) == 0

    def test_set_sequencenumber(self):
        "Sequence number is not defined by default but can be set manually"
        augmented_diff = AugmentedDiff()
        assert not augmented_diff.sequence_number
        augmented_diff.sequence_number = 12345
        assert augmented_diff.sequence_number == 12345
        augmented_diff.sequence_number = "12345"
        assert augmented_diff.sequence_number == 12345

    def test_read_changeset_from_xml_file(self, adiff_file_path):
        "Test initializing from an XML object"
        # not implemented yet
        pass



from osmdiff import Node
from osmdiff.osm import OSMObject
from typing_extensions import assert_type


class TestNode:
    "tests for Node object"

    def test_init_node(self):
        "Test Node init"
        node = Node()
        assert_type(node, Node)
        assert_type(node, OSMObject)
        assert_type(node.attribs, dict)
        assert_type(node.tags, dict)
        assert len(node.attribs) == 0
        assert len(node.tags) == 0
        assert node.lat == 0.0
        assert node.lon == 0.0



"""
osmdiff is a Python library for working with OpenStreetMap changesets and diffs.

It provides classes for working with OpenStreetMap changesets and diffs, and
includes a parser for the OpenStreetMap changeset API.
"""

from .augmenteddiff import AugmentedDiff
from .osm import Node, Relation, Way
from .osmchange import OSMChange

__version__ = "0.4.3"



from posixpath import join as urljoin
from gzip import GzipFile
from xml.etree import ElementTree

import requests

from osmdiff.osm import OSMObject
from osmdiff.settings import DEFAULT_REPLICATION_URL


class OSMChange(object):
    """
    Class to represent an OSMChange object.

    :param url: URL of the OSM replication server
    :type url: str
    :param frequency: frequency of the replication diff
    :type frequency: str
    :param file: path to the XML file
    :type file: str
    :param sequence_number: sequence number of the diff
    :type sequence_number: int
    """

    def __init__(
        self,
        url=DEFAULT_REPLICATION_URL,
        frequency="minute",
        file=None,
        sequence_number=None,
    ):
        self.base_url = url
        self.create = []
        self.modify = []
        self.delete = []
        if file:
            with open(file, "r") as fh:
                xml = ElementTree.iterparse(fh, events=("start", "end"))
                self._parse_xml(xml)
        else:
            self._frequency = frequency
            self._sequence_number = sequence_number

    def get_state(self) -> bool:
        """
        Get the current state from the OSM API.
        """
        # FIXME this should really not return a boolean
        """Get the current state from the OSM API"""
        state_url = urljoin(self.base_url, self._frequency, "state.txt")
        response = requests.get(state_url, timeout=5)
        if response.status_code != 200:
            return False
        for line in response.text.split("\n"):
            if line.startswith("sequenceNumber"):
                self._sequence_number = int(line[15:])
        return True

    def _build_sequence_url(self) -> str:
        seqno = str(self._sequence_number).zfill(9)
        url = urljoin(
            self.base_url,
            self._frequency,
            seqno[:3],
            seqno[3:6],
            "{}{}".format(seqno[6:], ".osc.gz"),
        )
        return url

    def _parse_xml(self, xml) -> None:
        for event, elem in xml:
            if elem.tag in ("create", "modify", "delete"):
                self._build_action(elem)

    def _build_action(self, elem) -> None:
        for thing in elem:
            o = OSMObject.from_xml(thing)
            self.__getattribute__(elem.tag).append(o)

    def retrieve(self, clear_cache=False, timeout=30) -> int:
        """
        Retrieve the OSM diff corresponding to the OSMChange sequence_number.

        :param clear_cache: clear the cache
        :type clear_cache: bool
        :param timeout: request timeout
        :type timeout: int

        :return: HTTP status code
        :rtype: int
        """
        if not self._sequence_number:
            raise Exception("invalid sequence number")
        if clear_cache:
            self.create, self.modify, self.delete = ([], [], [])
        try:
            r = requests.get(self._build_sequence_url(), stream=True, timeout=timeout)
            if r.status_code != 200:
                return r.status_code
            gzfile = GzipFile(fileobj=r.raw)
            xml = ElementTree.iterparse(gzfile, events=("start", "end"))
            self._parse_xml(xml)
            return r.status_code
        except ConnectionError:
            # FIXME catch this?
            return 0

    @classmethod
    def from_xml(cls, xml) -> "OSMChange":
        """
        Initialize OSMChange object from an XML object.

        If you used this method before version 0.3, please note that this
        method now takes an XML object. If you want to initialize from a file,\
        use the from_xml_file method.

        :param path: path to the XML file
        :type path: str

        :return: OSMChange object
        :rtype: OSMChange
        """
        new_osmchange_obj = cls()
        new_osmchange_obj._parse_xml(xml)
        return new_osmchange_obj

    @classmethod
    def from_xml_file(cls, path) -> "OSMChange":
        """
        Initialize OSMChange object from an XML file.

        :param path: path to the XML file
        :type path: str

        :return: OSMChange object
        :rtype: OSMChange
        """
        with open(path, "r") as fh:
            xml = ElementTree.iterparse(fh, events=("start", "end"))
            return cls.from_xml(xml)

    @property
    def sequence_number(self) -> int:
        return self._sequence_number

    @sequence_number.setter
    def sequence_number(self, value):
        try:
            # value can be none
            if value is None:
                self._sequence_number = None
                return
            self._sequence_number = int(value)
        except ValueError:
            raise ValueError(
                "sequence_number must be an integer or parsable as an integer"
            )

    @property
    def frequency(self) -> str:
        return self._frequency

    @frequency.setter
    def frequency(self, f) -> None:
        self._frequency = f

    def __repr__(self):
        return "OSMChange ({create} created, {modify} modified, \
{delete} deleted)".format(
            create=len(self.create), modify=len(self.modify), delete=len(self.delete)
        )



from osmdiff import Relation
from osmdiff.osm import OSMObject
from typing_extensions import assert_type


class TestRelation:
    "tests for Relation object"

    def test_init_relation(self):
        "Test Relation init"
        relation = Relation()
        assert_type(relation, Relation)
        assert_type(relation, OSMObject)
        assert_type(relation.attribs, dict)
        assert_type(relation.tags, dict)
        assert len(relation.tags) == 0
        assert len(relation.attribs) == 0
        assert_type(relation.members, list)
        assert len(relation.members) == 0



class OSMObject(object):
    def __init__(self) -> None:
        self.tags = {}
        self.attribs = {}
        self.bounds = None

    def __repr__(self) -> str:
        out = "{type} {id}".format(type=type(self).__name__, id=self.attribs.get("id"))
        if type(self) == Way:
            out += " ({ways} nodes)".format(ways=len(self.nodes))
        if type(self) == Relation:
            out += " ({mem} members)".format(mem=len(self.members))
        return out

    def _parse_tags(self, elem) -> None:
        for tagelem in elem.findall("tag"):
            self.tags[tagelem.attrib["k"]] = tagelem.attrib["v"]

    def _parse_bounds(self, elem) -> None:
        be = elem.find("bounds")
        if be is not None:
            self.bounds = [
                be.attrib["minlon"],
                be.attrib["minlat"],
                be.attrib["maxlon"],
                be.attrib["maxlat"],
            ]

    @classmethod
    def from_xml(cls, elem):
        osmtype = ""
        if elem.tag == "member":
            osmtype = elem.attrib["type"]
        else:
            osmtype = elem.tag
        if osmtype in ("node", "nd"):
            o = Node()
        elif osmtype == "way":
            o = Way()
            o._parse_nodes(elem)
        elif osmtype == "relation":
            o = Relation()
            o._parse_members(elem)
        else:
            pass
        o.attribs = elem.attrib
        o._parse_tags(elem)
        o._parse_bounds(elem)
        return o


class Node(OSMObject):
    def __init__(self):
        super().__init__()

    def _lon(self):
        return float(self.attribs.get("lon", 0))

    lon = property(_lon)

    def _lat(self):
        return float(self.attribs.get("lat", 0))

    lat = property(_lat)

    def _geo_interface(self):
        return {"type": "Point", "coordinates": [self.lon, self.lat]}

    __geo_interface__ = property(_geo_interface)

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        return self.lon == other.lon and self.lat == other.lat


class Way(OSMObject):
    def __init__(self):
        self.nodes = []
        super().__init__()

    def _parse_nodes(self, elem):
        for node in elem.findall("nd"):
            self.nodes.append(OSMObject.from_xml(node))

    def _geo_interface(self):
        geom_type = "LineString" if self.nodes[0] == self.nodes[-1] else "Polygon"
        return {
            "type": geom_type,
            "coordinates": [[[n.lon, n.lat] for n in self.nodes]],
        }

    __geo_interface__ = property(_geo_interface)


class Relation(OSMObject):
    def __init__(self):
        self.members = []
        super().__init__()

    def _parse_members(self, elem):
        for member in elem.findall("member"):
            self.members.append(OSMObject.from_xml(member))

    def _geo_interface(self):
        return {
            "type": "FeatureCollection",
            "Features": [[f.__geo_interface__ for f in self.members]],
        }

    __geo_interface__ = property(_geo_interface)



from posixpath import join as urljoin
from xml.etree import cElementTree

import dateutil.parser
import requests

from .osm import OSMObject

from osmdiff.settings import DEFAULT_OVERPASS_URL


class AugmentedDiff(object):
    """
    Class to represent an Augmented Diff object.
    """

    base_url = DEFAULT_OVERPASS_URL
    minlon = None
    minlat = None
    maxlon = None
    maxlat = None
    timestamp = None
    remarks = []

    def __init__(
        self,
        minlon=None,
        minlat=None,
        maxlon=None,
        maxlat=None,
        file=None,
        sequence_number=None,
        timestamp=None,
    ):
        self.create = []
        self.modify = []
        self.delete = []
        if file:
            with open(file, "r") as file_handle:
                self._parse_stream(file_handle)
        else:
            self.sequence_number = sequence_number
            if minlon and minlat and maxlon and maxlat:
                if maxlon > minlon and maxlat > minlat:
                    self.minlon = minlon
                    self.minlat = minlat
                    self.maxlon = maxlon
                    self.maxlat = maxlat
                else:
                    raise Exception("invalid bbox.")

    def get_state(self):
        """Get the current state from the OSM API"""
        state_url = urljoin(self.base_url, "augmented_diff_status")
        response = requests.get(state_url, timeout=5)
        if response.status_code != 200:
            return False
        self.sequence_number = int(response.text)
        return True

    def _build_adiff_url(self):
        url = "{base}/augmented_diff?id={sequence_number}".format(
            base=self.base_url, sequence_number=self.sequence_number
        )
        if self.minlon and self.minlat and self.maxlon and self.maxlat:
            url += "&bbox={minlon},{minlat},{maxlon},{maxlat}".format(
                minlon=self.minlon,
                minlat=self.minlat,
                maxlon=self.maxlon,
                maxlat=self.maxlat,
            )
        return url

    def _build_action(self, elem):
        if elem.attrib["type"] == "create":
            for child in elem:
                e = OSMObject.from_xml(child)
                self.__getattribute__("create").append(e)
        else:
            new = elem.find("new")
            old = elem.find("old")
            osm_obj_old = None
            osm_obj_new = None
            for child in old:
                osm_obj_old = OSMObject.from_xml(child)
            for child in new:
                osm_obj_new = OSMObject.from_xml(child)
            self.__getattribute__(elem.attrib["type"]).append(
                {"old": osm_obj_old, "new": osm_obj_new}
            )

    def _parse_stream(self, stream):
        for event, elem in cElementTree.iterparse(stream):
            if elem.tag == "remark":
                self.remarks.append(elem.text)
            if elem.tag == "meta":
                timestamp = dateutil.parser.parse(elem.attrib.get("osm_base"))
                self.timestamp = timestamp
            if elem.tag == "action":
                self._build_action(elem)

    def retrieve(self, clear_cache=False, timeout=30) -> int:
        """
        Retrieve the Augmented diff corresponding to the sequence_number.
        """
        if not self.sequence_number:
            raise Exception("invalid sequence number")
        if clear_cache:
            self.create, self.modify, self.delete = ([], [], [])
        url = self._build_adiff_url()
        try:
            r = requests.get(url, stream=True, timeout=timeout)
            if r.status_code != 200:
                return r.status_code
            r.raw.decode_content = True
            self._parse_stream(r.raw)
            return r.status_code
        except ConnectionError:
            # FIXME should we catch instead?
            return 0

    @property
    def remarks(self):
        return self._remarks

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        self._timestamp = value

    @property
    def sequence_number(self):
        return self._sequence_number

    @sequence_number.setter
    def sequence_number(self, value):
        try:
            # value can be none
            if value is None:
                self._sequence_number = None
                return
            self._sequence_number = int(value)
        except ValueError:
            raise ValueError(
                "sequence_number must be an integer or parsable as an integer"
            )

    def __repr__(self):
        return "AugmentedDiff ({create} created, {modify} modified, \
{delete} deleted)".format(
            create=len(self.create), modify=len(self.modify), delete=len(self.delete)
        )
